VERSION 1.0 CLASS
BEGIN
  MultiUse = -1  'True
  Persistable = 0  'NotPersistable
  DataBindingBehavior = 0  'vbNone
  DataSourceBehavior  = 0  'vbNone
  MTSTransactionMode  = 0  'NotAnMTSObject
END
Attribute VB_Name = "Stream"
Attribute VB_GlobalNameSpace = False
Attribute VB_Creatable = True
Attribute VB_PredeclaredId = False
Attribute VB_Exposed = True
'''' -------------------------
'''' Class TCPManagedStream v2
'''' -------------------------
'''' this class makes extensive use of the CopyMemory API
''''
'''' a class to handle the messages sent and received at both ends of a Winsock control's connection (TCP Mode).
'''' because TCP will combine small messages into one packet or split large messages over several packets
'''' for efficient transport across a network it is necessary to re-assemble the message back into its original
'''' state on receipt by the client
'''' This class will re-assemble the message before passing it onto the client via a DataArrival Event.
'
'''' the messages can be sent as either a string or Byte Array and the appropriate event will be fired on receipt
'
'''' For this class to work you must use both the Send and Receive methods, ie receiving messages that were not
'''' sent using this class's Send method will just be discarded and no data will be returned to the client.
'
'''' Usage
'''' -----
'''' Sending messages:-
'''' ==============
'''' Call the Stream.Send method with the appropriate arguments
'''' arg1 :- The Winsock Control that will send the Data
'''' arg2 :- the data to send (type specified by arg3)
'''' arg3 :- (optional) Datatype of the data (either a Byte array or String - defaults to byte array)
'''' arg4 :- (optional) a user defined identifier that can be used to distinguish different message types
'
'''' Receiving messages
'''' ==============
'''' Dim WithEvents a TCPManagedStream.Stream variable at the form level
'''' Implement the DataArrival Events Required
'
'''' Call the Stream.Receive Method inside the Winsock_DataArrival method
'''' arg1 :- the winsock control receiving the data, (can be part of a control array)
'
'''' Overhead
'''' --------
'''' for each call to the Send method a 6 byte header is added to the start of the client message

Option Explicit
Private Declare Sub CopyMemory Lib "kernel32" Alias "RtlMoveMemory" (Destination As Any, Source As Any, ByVal length As Long)
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''' Type of data to send
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Public Enum StreamDataType
  AsByteArray = 1
  AsString = 2
End Enum
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''' Type provides an array of buffered streams
''' 1 for each Winsock control in a control array
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Private Type Socket
  oWinsock                    As Winsock      ''' reference is passed back to the client in the DataArrival events
  vBuffer()                   As Byte         ''' holds the pre-processed data recieved by the winsock control
End Type
''' each item in the array corresponds to a winsock control array index
''' or index 0 is used if not part of a control array
Private m_sockets()           As Socket
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''' Header
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Private Const HEADER_SIZE         As Integer = 6
Private m_Header(HEADER_SIZE - 1) As Byte

Private Const TYPE_OFFSET         As Byte = 1       ''' offsets in the header for the individual values
Private Const IDENTIFIER_OFFSET   As Byte = 2
Private Const LENGTH_OFFSET       As Byte = 3
'
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
''' Events
''''''''''''''''''''''''''''''''''''''''''''''''''''''''''''
'
Event DataArrivalAsByteArray(vWinsock As Object, vData() As Byte, ByVal Identifier As Byte)
Event DataArrivalAsByteString(vWinsock As Object, vData As String, ByVal Identifier As Byte)

'---------------------------------------------------------------------------------------
' Procedure : Sub Receive
'------------------------
' the receive method gets any data from the supplied winsock control and adds it to
' a byte array buffer that is unique to that winsock control (it could be part of a control array)
' it then passes the data onto the ProcessMessageBuffer method.
'---------------------------------------------------------------------------------------
Public Sub Receive(vWinsock As Object)
  If vWinsock Is Nothing Then
    Err.Raise vbObjectError + 6342, "TCPManagedStream.Stream", "Winsock Error"
    Exit Sub
  End If
  
  Dim winsockIndex  As Integer
  Dim chunkSize     As Long
  Dim vData()       As Byte
  Dim bufferSize    As Integer

  vWinsock.GetData vData, vbByte Or vbArray   '' data always received as a byte array
  
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''' Size of the received chunk of data
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  On Error Resume Next
  chunkSize = UBound(vData) + 1
  If Err <> 0 Then Exit Sub
  If chunkSize < HEADER_SIZE Then Exit Sub  ''' must be at least the header in a chunk
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  '' Get the index of the Winsock
  '' if it's not part of a control array then use index 0
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  winsockIndex = vWinsock.Index
  If Err <> 0 Then winsockIndex = 0
  On Error GoTo 0
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''' add a new Socket to the array if needed
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  If winsockIndex > UBound(m_sockets) Then
    ReDim Preserve m_sockets(winsockIndex)
  End If
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''' add the data to the socket's buffer
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  With m_sockets(winsockIndex)
    If .oWinsock Is Nothing Then        ''' first data packet received for this socket
      Set .oWinsock = vWinsock
      ReDim .vBuffer(chunkSize - 1)
      CopyMemory .vBuffer(0), vData(0), chunkSize
    Else
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' add this chunk to the currently buffered data for this socket
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      bufferSize = UBound(.vBuffer) + 1  '' get the current size of the buffer
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' redim the socket's buffer preserving the current contents
      ''' make it large enough to hold the current contents and the new data
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ReDim Preserve .vBuffer(bufferSize + chunkSize - 1)
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' copy the new data into the socket's buffer after the original data
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      CopyMemory .vBuffer(bufferSize), vData(0), chunkSize
    End If
  End With
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  '''  SECTION THAT WILL PROCESS THE DATA STREAM
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  Dim streamType      As StreamDataType
  Dim messageLength   As Integer
  Dim Identifier      As Byte
  Dim bytesLeft       As Integer
  Dim s               As String
  Dim clientData()    As Byte
  Dim temp()          As Byte
    
  ' a message will consist of the 6 byte header and the client data
  ' the header contains :
  ' the message datatype      (1 byte)
  ' a user defined identifier (1 byte)
  ' the message length        (2 bytes)
  ' plus opening and closing tag markers
  
  ' once a complete message is in the buffer (ie the buffer size is greater than the header+data size)
  ' it is extracted from the buffer and sent to the client, the buffer is resized, removing the
  ' message and the process is repeated, once there are insufficient bytes left in the buffer, control is returned
  ' to the client and we wait for more data.

  With m_sockets(winsockIndex)
    Do ''' Loop until there are no complete messages in the buffer.
      
      ' the first 6 bytes in the socket's buffer will always be a message header (validity check)
      If .vBuffer(0) <> m_Header(0) Or .vBuffer(HEADER_SIZE - 1) <> m_Header(HEADER_SIZE - 1) Then
        Set .oWinsock = Nothing
        Erase .vBuffer
        Exit Do         ''' first exit point, invalid header markers, erase the rest of the buffer and return
      End If

      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' get the message length, stored at byte offset 'LENGTH_OFFSET' in the header
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      If (UBound(.vBuffer) + 1) >= HEADER_SIZE Then
        CopyMemory messageLength, .vBuffer(LENGTH_OFFSET), 2
      End If
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' check if there is enough bytes in the socket's buffer to hold the first message and it's header
      ''' if there isn't enough bytes in the socket buffer, return and wait for the next packet to arrive
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      If (HEADER_SIZE + messageLength > UBound(.vBuffer) + 1) Then
        Exit Do         ''' second exit point, not enough bytes in the buffer for a complete message
      End If
      
      
      '*****************************************************************
      '*****************************************************************
      ''' IF WE GET HERE THEN THERE IS A COMPLETE MESSAGE IN THE BUFFER
      '*****************************************************************
      '*****************************************************************
      streamType = .vBuffer(TYPE_OFFSET)          ''' get the stream type, stored at byte offset 'TYPE_OFFSET' in the header
      Identifier = .vBuffer(IDENTIFIER_OFFSET)    ''' get the user supplied value that can be used as a message type identifier
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' create a new byte array to hold the message
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      If messageLength > 0 Then
        ReDim clientData(messageLength - 1)
        CopyMemory clientData(0), .vBuffer(HEADER_SIZE), messageLength    '' copy the message minus the header
      Else
        Erase clientData   ''' return an empty array
      End If
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' raise an event back to the client with the completed message
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      If streamType = AsByteArray Then
        RaiseEvent DataArrivalAsByteArray(.oWinsock, clientData, Identifier)
      ElseIf streamType = AsString Then
        s = StrConv(clientData, vbUnicode)
        RaiseEvent DataArrivalAsByteString(.oWinsock, s, Identifier)
      End If
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      ''' check for any more messages in the bufffer
      ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
      bytesLeft = (UBound(.vBuffer) + 1) - HEADER_SIZE - messageLength
      If bytesLeft > 0 Then
        '''''''''''''''''''''''''''''''''''''''''''''''''''''''
        ''' resize the socket's buffer, removing the processed message
        '''''''''''''''''''''''''''''''''''''''''''''''''''''''
        ReDim temp(bytesLeft - 1)
        
        CopyMemory temp(0), .vBuffer(HEADER_SIZE + messageLength), bytesLeft
        .vBuffer = temp
      Else
        ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
        ''' all expected data has arrived so there are no more messages to process
        ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
        Set .oWinsock = Nothing
        Erase .vBuffer
        Exit Do                    ''' third exit point, all bytes in the stream processed
      End If
    Loop
  End With
End Sub

'---------------------------------------------------------------------------------------
' Procedure : Sub Send
'---------------------
' identifier is a user defined value that can be used to identify differing message types on receipt
'---------------------------------------------------------------------------------------
Public Sub Send(vWinsock As Object, _
                Optional vData As Variant, _
                Optional ByVal streamType As StreamDataType = AsByteArray, _
                Optional ByVal Identifier As Byte)
                
  If vWinsock Is Nothing Then
    Err.Raise vbObjectError + 6342, "TCPManagedStream.Stream", "Winsock Error"
    Exit Sub
  End If
  
  Dim buffer()  As Byte
  Dim work()    As Byte
  Dim length    As Integer
  ''' Check the stream Length
  ''' as we are using 2 bytes in the header to record the stream length
  ''' we are restricted to a maximum of 32767 bytes in the data stream
  ''' to increase this you can increase the header size by 2 bytes
  ''' and use a 4 bytes long integer value for the message length
  If streamType = AsByteArray Then
    On Error Resume Next
    length = UBound(vData) + 1   ''' could be nothing
    On Error GoTo 0
    
    If length > 32767 Then
      Err.Raise vbObjectError + 6345, "TCPManagedStream.Stream", "Maximum data length allowed per message is 32k"
      Exit Sub
    End If
  ElseIf streamType = AsString Then
    On Error Resume Next
    length = Len(vData)
    On Error GoTo 0
    
    If length > 32767 Then
      Err.Raise vbObjectError + 6345, "TCPManagedStream.Stream", "Maximum data length allowed per message is 32k"
      Exit Sub
    End If
  End If
  
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''' convert the stream to a byte array (if not already)
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  If length > 0 Then
    If streamType = AsByteArray Then
      work = vData
    ElseIf streamType = AsString Then
      work = StrConv(vData, vbFromUnicode)
    Else
      Err.Raise vbObjectError + 6343, "TCPManagedStream.Stream", "Invalid Data Type"
      Exit Sub
    End If
  End If
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ''' create a new byte array big enough for the header and data
  ''''''''''''''''''''''''''''''''''''''''''''''''''''''''
  ReDim buffer((length + HEADER_SIZE) - 1)
  
  CopyMemory buffer(0), m_Header(0), HEADER_SIZE                '' copy the header template
  buffer(TYPE_OFFSET) = CByte(streamType)                       '' copy the stream type
  buffer(IDENTIFIER_OFFSET) = Identifier                        '' copy the message identifier
  CopyMemory buffer(LENGTH_OFFSET), length, 2                   '' copy the stream length
  
  If length > 0 Then
    CopyMemory buffer(HEADER_SIZE), work(0), length             '' copy the actual client data
  End If
  
  ''' send it
  If vWinsock.State = sckConnected Then
    vWinsock.SendData buffer: DoEvents
  End If
End Sub

'---------------------------------------------------------------------------------------
' Procedure : Sub Class_Initialize
'---------------------------------------------------------------------------------------
Private Sub Class_Initialize()
  ''' this is the header template which is used in the receive method
  ''' to ensure that the data recieved is a valid message sent by this classes Send method
  
  m_Header(0) = 60    ' "<"
  m_Header(1) = 0     ' 1 byte  reservered for the stream type
  m_Header(2) = 0     ' 1 byte  reserved for a user defined message identifier
  m_Header(3) = 0     ' 2 bytes reservered for the stream length
  m_Header(4) = 0
  m_Header(5) = 62    ' ">"
  
  ''' create an initial array item
  ''' this index will be used if the winsock control
  ''' receiving the data is not part of a control array
  ReDim m_sockets(0)  '
End Sub

'---------------------------------------------------------------------------------------
' Procedure : Sub Class_Terminate
'---------------------------------------------------------------------------------------
Private Sub Class_Terminate()
  Erase m_sockets
End Sub
